@@ -21,6 +21,9 @@ from enum import Enum
 from numerous_api_client.headers.ValidationInterceptor import ValidationInterceptor
 import traceback
 from uuid import uuid4
+from dirtools import Dir, DirState, filehash
+from pathlib import Path
+
 
 class JobStatus(Enum):
     ready=0
@@ -167,35 +170,67 @@ class NumerousBufferedWriter:
 
 
 class NumerousLogHandler:
-    def __init__(self, client):
+    def __init__(self, client, admin_log_id=None, echo=False):
         self.client = client
         self.buffered_records = []
         self.last_push = datetime.utcnow() - timedelta(hours=1)
         self.debounce = 5
+        self.admin_log_id = admin_log_id
+        self.echo = echo
+
+        def timed_flush():
+            while True:
+                sleep(1)
+                now_ = datetime.utcnow()
+                if (now_ - self.last_push).seconds > self.debounce:
+                    self.last_push = now_
+                    self.flush()
+        threading.Thread(target=timed_flush, daemon=True).start()
+
 
     def handle(self, record):
+        self.log(record.msg)
 
+
+    def log(self, message):
         now_ = datetime.utcnow()
-        #print(record)
-        self.buffered_records.append((now_, record.msg))
+        # print(record)
+        record = (now_, message)
+        self.buffered_records.append(record)
+        if self.echo:
+            print(record)
 
         if (now_ - self.last_push).seconds > self.debounce:
             self.last_push = now_
             self.flush()
 
-
     def flush(self):
+        if len(self.buffered_records)>0:
+            to_push = self.buffered_records.copy()
+            self.buffered_records = []
+            if self.admin_log_id is not None:
+                self.client.push_scenario_logs_admin(to_push, exe_id=self.admin_log_id)
+            else:
+                self.client.push_scenario_logs(to_push)
 
-        to_push = self.buffered_records.copy()
-        self.buffered_records = []
-        self.client.push_scenario_logs(to_push)
+    def close(self):
+        self.flush()
 
 
     # TODO make a handler thread to schedule update to server!
 
 
 class NumerousClient:
-    def __init__(self, job_id=None, project=None, scenario=None, server=None, port=None, refresh_token=None, clear_data=None, no_log=False, instance_id=None, secure=None):
+    def __init__(self, job_id=None, project=None, scenario=None, server=None, port=None, refresh_token=None, clear_data=None, no_log=False, instance_id=None, secure=None, url=None):
+
+        if url is not None:
+            from urllib.parse import urlparse
+            parsed_url = urlparse(url)
+
+            server=parsed_url.netloc.split(':')[0]
+            port=parsed_url.netloc.split(':')[1]
+            secure=parsed_url.scheme == "https"
+
         self.job_states = JobStatus
         self._project = get_env(project, 'NUMEROUS_PROJECT')
         self._scenario = get_env(scenario, 'NUMEROUS_SCENARIO')
@@ -205,9 +240,10 @@ class NumerousClient:
 
         self._instance_id = str(uuid4()) if instance_id is None else instance_id
         self._execution_id = get_env(None, 'NUMEROUS_EXECUTION_ID', 'not_found')
+        print('Exe: ',self._execution_id)
         if self._execution_id == 'not_found' or self._execution_id is None:
-            self._execution_id = str(uuid4())
-            log.debug(f'No execution id found - generating new: {self._execution_id}')
+            self._execution_id = str(uuid4()).replace('-','_')
+            print(f'No execution id found - generating new: {self._execution_id}')
 
         self._complete = True
         self.writers = []
@@ -375,17 +411,38 @@ class NumerousClient:
     #        scenario = self._scenario
 
     #    self._base_client.SetScenarioDataTags(spm_pb2.ScenarioDataTags(project=self._project, scenario=scenario, tags=tags))
-
-    def set_scenario_progress(self, message, status, progress=None, clean=False, scenario=None, force=False):
+    def set_scenario_job_image(self, name, path, project=None, scenario=None, job=None):
         if scenario is None:
             scenario = self._scenario
+
+        if project is None:
+            project = self._project
+
+        if job is None:
+            job = self._job_id
+
+        self._base_client.SetScenarioJobImage(spm_pb2.ScenarioJobImage(
+            project_id=project, scenario_id=scenario, job_id=job,
+            name=name, path=path
+        ))
+
+    def set_scenario_progress(self, message, status, progress=None, clean=False, scenario=None, force=False, project=None, job=None):
+        if scenario is None:
+            scenario = self._scenario
+
+        if project is None:
+            project = self._project
+
+        if job is None:
+            job = self._job_id
+
         tic = time()
 
         if tic > (self.last_progress + self.progress_debounce) or force:
 
             self.last_progress=tic
             self._base_client.SetScenarioProgress(spm_pb2.ScenarioProgress(
-                project=self._project, scenario=scenario, job_id=self._job_id,
+                project=project, scenario=scenario, job_id=job,
                 message=message, status=status, clean=clean, progress=progress
             ))
 
@@ -487,19 +544,28 @@ class NumerousClient:
 
         upload()
 
-    def get_timeseries_stats(self, scenario=None):
+    def get_timeseries_stats(self, project=None, scenario=None, execution=None):
+        if project is None:
+            project = self._project
+        if scenario is None:
+            scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
+
+        return self._base_client.GetScenarioDataStats(spm_pb2.Scenario(project=project, scenario=scenario, execution=execution))
+
+    def get_timeseries_meta_data(self, project=None, scenario=None, execution=None):
+        if project is None:
+            project = self._project
         if scenario is None:
             scenario = self._scenario
 
-        return self._base_client.GetScenarioDataStats(spm_pb2.Scenario(scenario=scenario))
 
-    def get_timeseries_meta_data(self, scenario=None):
-        if scenario is None:
-            scenario = self._scenario
 
-        return self._base_client.GetScenarioMetaData(spm_pb2.Scenario(scenario=scenario))
 
-    def set_timeseries_meta_data(self, tags:list, aliases:dict={}, offset=0, timezone='UTC', epoch_type='s', scenario= None):
+        return self._base_client.GetScenarioMetaData(spm_pb2.Scenario(project=project, scenario=scenario, execution=execution))
+
+    def set_timeseries_meta_data(self, tags:list, aliases:dict={}, offset=0, timezone='UTC', epoch_type='s', scenario= None, execution=None):
 
         def g(tag, key, def_):
 
@@ -528,54 +594,98 @@ class NumerousClient:
 
         if scenario is None:
             scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
         return self._base_client.SetScenarioMetaData(
-            spm_pb2.ScenarioMetaData(scenario=scenario,
+            spm_pb2.ScenarioMetaData(scenario=scenario, execution=execution,
             tags=[wrap_tag(tag) for tag in tags],
             aliases=[spm_pb2.Alias(tag=k, aliases=v) for k,v in aliases.items()],
                                      offset=offset, timezone=timezone, epoch_type=epoch_type
                             )
         )
 
-    def get_timeseries_custom_meta_data(self, scenario=None, key=None):
+    def get_timeseries_custom_meta_data(self, scenario=None, execution=None, key=None):
         if scenario is None:
             scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
-        return self._base_client.GetScenarioCustomMetaData(spm_pb2.ScenarioCustomMetaData(scenario=scenario, key=key)).meta
+        return self._base_client.GetScenarioCustomMetaData(spm_pb2.ScenarioCustomMetaData(scenario=scenario, execution=execution, key=key)).meta
 
-    def set_timeseries_custom_meta_data(self, meta:str, scenario=None, key=None):
+    def set_timeseries_custom_meta_data(self, meta:str, scenario=None, execution=None, key=None):
         if scenario is None:
             scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
-        return self._base_client.SetScenarioCustomMetaData(spm_pb2.ScenarioCustomMetaData(scenario=scenario, key=key, meta=meta))
+        return self._base_client.SetScenarioCustomMetaData(spm_pb2.ScenarioCustomMetaData(scenario=scenario, execution=execution, key=key, meta=meta))
+
+    def get_state(self, scenario=None, execution=None):
+        if scenario is None:
+            scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
-    def get_state(self, scenario=None):
         try:
-            state_json = self.get_timeseries_custom_meta_data(scenario=None, key='state')
+            state_json = self.get_timeseries_custom_meta_data(scenario=None, execution=execution, key='state')
             return json.loads(state_json)
         except json.decoder.JSONDecodeError:
             #print('bad json: ', state_json)
             return None
 
-    def set_state(self, state, scenario=None, force=False):
+    def set_state(self, state, scenario=None, execution=None, force=False):
         tic = time()
+        if scenario is None:
+            scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
         if tic > (self.last_state_set + self.state_debounce) or force:
             log.debug('State set.')
             self.last_state_set = tic
-            return self.set_timeseries_custom_meta_data(meta=json.dumps(state),scenario=scenario, key='state')
+            return self.set_timeseries_custom_meta_data(meta=json.dumps(state),scenario=scenario, execution=execution, key='state')
 
-    def data_read_df(self, tags=[], start:datetime=None, end:datetime=None, scenario=None):
-        if not scenario:
+    def get_latest_main_execution(self, project=None, scenario=None):
+        return self._base_client.GetLatestMainExecution(spm_pb2.Scenario(project=project, scenario=scenario)).execution_id
+
+
+    def data_read_df(self, tags=[], start:datetime=None, end:datetime=None, project=None, scenario=None, execution=None):
+        #print(tags)
+
+        if project is None:
+            project = self._project
+
+        if scenario is None:
             scenario = self._scenario
 
+        if execution is None:
+            execution = self.get_latest_main_execution(project, scenario)
+            log.debug('Exe: ' + execution)
+
+        #if execution is None:
+        #    execution = self._execution_id
+
+        if  start is None:
+            start=0
+        if end is None:
+            end = 0
+
+        time_range = (start > 0 or end > 0)
+
+
         read_data = self._base_client.ReadData(
-            spm_pb2.ReadScenario(scenario=scenario, tags=tags, start=start, end=start))
+            spm_pb2.ReadScenario(scenario=scenario, execution=execution, tags=tags, start=start, end=end, time_range=time_range,
+                                 listen=False))
+
+        #read_data = self._base_client.ReadData(
+        #    spm_pb2.ReadScenario(scenario=scenario, tags=tags, start=start, end=end, time_range=time_range, listen=False))
 
         #data = {r.tag: r.values for l in read_data for r in l.data}
         data = {}
         min_len = 1e12
         for l in read_data:
+            #print('l: ', l)
 
             for r in l.data:
                 if not r.tag in data:
@@ -597,7 +707,7 @@ class NumerousClient:
 
         return df
 
-    def read_data_stream(self, tags, start:datetime=None, end:datetime=None, scenario=None):
+    def read_data_stream(self, tags, start:datetime=None, end:datetime=None, scenario=None, execution=None):
 
         time_range= False
         if isinstance(start, datetime):
@@ -621,32 +731,52 @@ class NumerousClient:
         if not scenario:
             scenario = self._scenario
 
-        read_data = self._base_client.ReadData(spm_pb2.ReadScenario(scenario=scenario, tags=tags, start=start, end=end, time_range = time_range))
+        if execution is None:
+            execution = self._execution_id
+
+        read_data = self._base_client.ReadData(spm_pb2.ReadScenario(scenario=scenario, execution=execution, tags=tags, start=start, end=end, time_range = time_range))
 
         for l in read_data:
             for r in l.data:
                 yield r.tag, r.values
 
-    def read_data_as_row(self, tags=[], start:int=0, end:int=0, scenario=None, subscribe=False, offset=0):
+    def read_data_as_row(self, tags=[], start:int=0, end:int=0, scenario=None, execution=None, subscribe=False, offset=0, project=None):
+        if project is None:
+            project = self._project
 
+        print(end)
         log.debug('Reading data as row')
         log.debug('Subscribed: ' + str(subscribe) + ', starting from: '+str(start))
         if not scenario:
             scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
+
         time_range = (start > 0 or end > 0)
+
+        end = end if end > 0 else 3600*24*365*10000
+
+        print(start)
+        print(end)
+
         read_data = self._base_client.ReadData(
-            spm_pb2.ReadScenario(scenario=scenario, tags=tags, start=start, end=end, time_range=time_range, listen=subscribe))
+            spm_pb2.ReadScenario(project=project, scenario=scenario, execution=execution, tags=tags, start=start, end=end, time_range=time_range, listen=subscribe))
         current_row = {}
-        end = end if end>0 else 9999999
+
+
         for b in read_data:
 
-            for d in b.data:
-                if not d.tag in current_row:
-                    current_row[d.tag]=list(d.values)
-                else:
-                    current_row[d.tag] += list(d.values)
 
-            if b.row_complete:
+            for d in b.data:
+                if d.tag != "":
+                    if not d.tag in current_row:
+                        current_row[d.tag]=list(d.values)
+                    else:
+                        current_row[d.tag] += list(d.values)
+
+            if b.row_complete or b.block_complete:
+                print(current_row.keys())
+                print(tags)
                 len_index = len(current_row['_index'])
 
                 wrong_len_tags = []
@@ -662,13 +792,18 @@ class NumerousClient:
                     if current_ix >= start and current_ix < end:
 
                         r = {k: v[i] for k,v in current_row.items()}
-                        r['_index']+=offset
-                        yield r
+                        if not time_range or (r['_index']>=start and r['_index']<end):
+                            r['_index']+=offset
+
+                            yield r
+
 
                 current_row = {}
 
         if len(list(current_row.keys()))>0:
+
             for i, ix in enumerate(current_row['_index']):
+                print(i)
                 r = {k: v[i] for k, v in current_row.items()}
                 r['_index'] += offset
                 yield r
@@ -677,21 +812,25 @@ class NumerousClient:
         log.debug('Reading as row complete')
         return
 
-    def clear_timeseries_data(self, scenario=None):
+    def clear_timeseries_data(self, scenario=None, execution=None):
         log.warning('Clearing data!')
 
         if scenario is None:
             scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
         self._base_client.ClearData(
-            spm_pb2.Scenario(scenario=scenario))
+            spm_pb2.Scenario(scenario=scenario, execution=execution))
 
-    def data_write_df(self, df, scenario=None, clear=True):
+    def data_write_df(self, df, scenario=None, execution=None, clear=True):
         if scenario is None:
             scenario = self._scenario
+        if execution is None:
+            execution = self._execution_id
 
         if clear:
-            self.clear_timeseries_data(scenario=scenario)
+            self.clear_timeseries_data(scenario=scenario, execution=execution)
 
         #df['_index']=[t.timestamp() for t in df.index]
         len_ix = len(df['_index'])
@@ -712,11 +851,11 @@ class NumerousClient:
                     size += b_size
                     blocks.append(spm_pb2.DataBlock(**d))
                     if size>2e5:
-                        yield spm_pb2.DataList(scenario=scenario, data=blocks, clear=False, block_complete=False, row_complete=False)
+                        yield spm_pb2.DataList(scenario=scenario, execution=execution, data=blocks, clear=False, block_complete=False, row_complete=False)
                         size=0
                         blocks = []
                 if size>0:
-                    yield spm_pb2.DataList(scenario=scenario, data=blocks, clear=False, block_complete=True,
+                    yield spm_pb2.DataList(scenario=scenario, execution=execution, data=blocks, clear=False, block_complete=True,
                                            row_complete=False)
 
         tic = time()
@@ -741,7 +880,7 @@ class NumerousClient:
 
         return blocks
 
-    def write_with_row_generator(self, data_generator,  scenario:str=None, clear = False, must_flush=None):
+    def write_with_row_generator(self, data_generator,  scenario:str=None, clear = False, must_flush=None, execution=None):
         if must_flush is None:
             def must_flush_(n):
                 return True
@@ -751,6 +890,8 @@ class NumerousClient:
         if scenario is None:
             scenario = self._scenario
 
+        if execution is None:
+            execution = self._execution_id
 
         def gen(clear):
 
@@ -763,14 +904,14 @@ class NumerousClient:
                     blocks = self._rows_to_blocks(rows)
                     rows = []
 
-                    yield spm_pb2.DataList(scenario=scenario, data=blocks, clear=clear, block_complete=False, row_complete=True)
+                    yield spm_pb2.DataList(scenario=scenario, execution=execution, data=blocks, clear=clear, block_complete=False, row_complete=True)
                     clear = False
 
             #if len(rows)>0:
             blocks = self._rows_to_blocks(rows)
 
 
-            yield spm_pb2.DataList(scenario=scenario, data=blocks, clear=clear, block_complete=True, row_complete=True)
+            yield spm_pb2.DataList(scenario=scenario, execution=execution, data=blocks, clear=clear, block_complete=True, row_complete=True)
 
             #return
 
@@ -806,11 +947,13 @@ class NumerousClient:
         finally:
             writer_.close()
 
-    def get_inputs(self, scenario_data, t0=0, dt=3600, tag_prefix='Stub', tag_seperator='.', timeout=10, subscribe=False):
+    def get_inputs(self, scenario_data, t0=0, te=0, dt=3600, tag_prefix='Stub', tag_seperator='.', timeout=10, subscribe=False):
         input_sources = {}
         static_data = {}
         input_source_types = {}
         only_static = True
+        input_scenario_map_spec = {s['scenarioID']: s for s in
+                                   scenario_data['inputScenarios']}
 
         for c in scenario_data['simComponents']:
             if "subcomponents" in c:
@@ -835,13 +978,15 @@ class NumerousClient:
                         only_static = False
 
                         if iv['dataSourceID'] not in input_sources:
-                            input_sources[iv['dataSourceID']] = {}
+                            input_sources[iv['dataSourceID']] = input_scenario_map_spec[iv['dataSourceID']]
+                            input_sources[iv['dataSourceID']]['tags'] = {}
 
-                        input_sources[iv['dataSourceID']][tag_] = {'tag': iv['tagSource']['tag'], 'scale': iv['scaling'],
+                        input_sources[iv['dataSourceID']]['tags'][tag_] = {'tag': iv['tagSource']['tag'], 'scale': iv['scaling'],
                                                                  'offset': iv['offset']}
                         input_source_types[iv['dataSourceID']]='scenario'
 
-                    elif iv['dataSourceType'] in ['dataset','csv']:
+                        """
+                        elif iv['dataSourceType'] in ['dataset','csv']:
 
                         only_static = False
 
@@ -851,7 +996,7 @@ class NumerousClient:
                         input_sources[iv['dataSourceID']][tag_] = {'tag': iv['tagSource']['tag'], 'scale': iv['scaling'],
                                                                    'offset': iv['offset']}
                         input_source_types[iv['dataSourceID']]='csv'
-
+                        """
                     else:
                         raise ValueError('Unknown data source type: ' + iv['dataSourceType'])
 
@@ -860,16 +1005,22 @@ class NumerousClient:
         inputs = [StaticSource(static_data, t0=t0, dt=dt, generate_index=only_static)]
         log.debug('input source types: ' + str(input_source_types))
         # Then add dynamic sources
+        input_scenario_map = {s['scenarioID']: s['executionID'] if 'executionID' in s else None for s in scenario_data['inputScenarios']}
+
+
         for input_source, spec in input_sources.items():
-            inputs.append(DynamicSource(client=self, scenario=input_source, spec=spec, t0=t0, timeout=timeout, source_type=input_source_types[input_source], subscribe=subscribe))
+            log.debug(spec)
+            log.debug('Setting up input sources, scenario: '+input_source + " exe:"+ input_scenario_map[input_source])
+
+            inputs.append(DynamicSource(client=self, scenario=input_source, execution=input_scenario_map[input_source], spec=spec, t0=t0, te=te, timeout=timeout, source_type=input_source_types[input_source], subscribe=subscribe))
 
         input_manager = InputManager(inputs)
         return input_manager
 
-    def iter_inputs(self, scenario_data, t0=0, dt=3600, tag_prefix='Stub', tag_seperator='.', timeout=10, subscribe=False):
+    def iter_inputs(self, scenario_data, t0=0, te=0,  dt=3600, tag_prefix='Stub', tag_seperator='.', timeout=10, subscribe=False):
 
         while True:
-            for r in self.get_inputs(scenario_data, t0=t0, dt=dt, tag_prefix=tag_prefix, tag_seperator=tag_seperator, timeout=timeout, subscribe=subscribe):
+            for r in self.get_inputs(scenario_data, t0=t0, te=te, dt=dt, tag_prefix=tag_prefix, tag_seperator=tag_seperator, timeout=timeout, subscribe=subscribe):
                 yield r
             return
 
@@ -900,7 +1051,7 @@ class NumerousClient:
 
         if hasattr(self, 'log_handler'):
             log.debug('flushing log')
-            self.log_handler.flush()
+            self.log_handler.close()
 
         log.debug('Client closing')
         if hasattr(self,'log_listener'):
@@ -944,8 +1095,10 @@ class StaticSource:
         return self.row
 
 class DynamicSource:
-    def __init__(self, client, scenario, spec, t0=0, timeout=None, source_type=None, subscribe=False):
+    def __init__(self, client, scenario, execution, spec, t0=0, te=0, timeout=None, source_type=None, subscribe=False):
         self.timeout = timeout
+        print('Scenario: ', scenario)
+        print('t0: ', 10)
 
         if source_type == 'csv':
             t0/=3600
@@ -954,13 +1107,31 @@ class DynamicSource:
             log.debug('Continous mode set on input source')
         else:
             continuous = False
+        self.meta = client.get_timeseries_meta_data(scenario=scenario, execution=execution)
+        print(spec)
 
-        self.generator = client.read_data_as_row(tags=[s['tag'] for s in spec.values()], scenario=scenario, start=t0, subscribe=subscribe)
         self.spec = spec
+
+        self.spec['offset'] = self.spec['offset'] if 'offset' in self.spec else 0
+
+        print('t0: ', datetime.fromtimestamp(t0))
+        print('te: ', datetime.fromtimestamp(te))
+        print('spec off: ', datetime.fromtimestamp(self.spec['offset']))
+
+        print('meta off: ', datetime.fromtimestamp(self.meta.offset))
+        print('start: ', datetime.fromtimestamp(t0-self.spec['offset']))
+        print('end: ', datetime.fromtimestamp(te-self.spec['offset']))
+
+
+
+        self.generator = client.read_data_as_row(tags=[s['tag'] for s in spec['tags'].values()], scenario=scenario,
+                                                 execution=execution, start=t0 - self.spec['offset'], end=te-self.spec['offset'],
+                                                 subscribe=subscribe)
         self.source_type = source_type
         self.row_queue = Queue(100)
 
         def _next():
+
             offset = 0
             while True:
                 if not self.row_queue.full():
@@ -968,8 +1139,9 @@ class DynamicSource:
                     while continuous or first:
                         first = False
                         try:
+
                             row_ = self.generator.__next__()
-                            #print(row_['_index'])
+
                             offset = row_['_index']
                             self.row_queue.put(row_)
 
@@ -981,7 +1153,7 @@ class DynamicSource:
                             else:
                                 log.debug('Requerying data')
                                 self.generator = client.read_data_as_row(tags=[s['tag'] for s in self.spec.values()],
-                                                                         scenario=scenario, start=0,
+                                                                         scenario=scenario, execution=execution, start=self.spec['offset'],
                                                                          subscribe=subscribe, offset=offset)
 
                 else:
@@ -992,6 +1164,7 @@ class DynamicSource:
         threading.Thread(target=_next, daemon=True).start()
 
     def __next__(self):
+
         try:
 
             row_ = self.row_queue.get(timeout=self.timeout)
@@ -1004,9 +1177,12 @@ class DynamicSource:
             raise StopIteration('Iterator consumed')
 
         row = {}
-        for tag, s in self.spec.items():
+        for tag, s in self.spec['tags'].items():
             row[tag] = row_[s['tag']]*s['scale'] + s['offset']
         row['_index'] = row_['_index']
+        row['_index_offset'] = row_['_index'] +self.meta.offset + self.spec['offset']
+        row['_datetime_utc'] = datetime.utcfromtimestamp(row['_index_offset'])
+        #row['_datatime_localized'] = datetime.fromtimestamp(row['_index_offset'], tz=datetime.timetz(self.meta.timezone))
 
         return row
 
@@ -1015,8 +1191,10 @@ class InputManager:
         self.sources = sources
 
     def get_data_from_sources(self):
+
         row = {}
         for source in self.sources:
+
             part_row = source.__next__()
             if isinstance(part_row, dict):
                 if '_index' in row and '_index' in part_row and row['_index'] != part_row['_index']:
@@ -1035,7 +1213,8 @@ class InputManager:
         while True:
 
             try:
-              row = self.get_data_from_sources()
+
+                row = self.get_data_from_sources()
             except StopIteration:
 
                 return
@@ -1078,11 +1257,14 @@ class Test(NumerousClient):
 
 
 class NumerousAdminClient(NumerousClient):
-    def __init__(self, server=None, port=None, secure=None, refresh_token=None, instance_id=None):
+    def __init__(self, server=None, port=None, secure=None, refresh_token=None, instance_id=None, url=None):
         super(NumerousAdminClient, self).__init__(
-            server=server, port=port, secure=secure, refresh_token=refresh_token, instance_id=instance_id, project="", scenario="", job_id=""
+            server=server, port=port, secure=secure, refresh_token=refresh_token, instance_id=instance_id, project="", scenario="", job_id="", url=url
         )
         self._job_manager = spm_pb2_grpc.JobManagerStub(self.channel)
+        self._build_manager = spm_pb2_grpc.BuildManagerStub(self.channel)
+        self._token_manager = spm_pb2_grpc.TokenManagerStub(self.channel)
+
         self._terminated = False
         self._complete = False
         self._user = 'test_user'
@@ -1115,6 +1297,12 @@ class NumerousAdminClient(NumerousClient):
     def get_execution_status(self, exe_id):
         return self._job_manager.GetExecutionStatus(spm_pb2.ExecutionId(execution_id=exe_id)).json
 
+    def clear_execution(self, scenario, execution):
+        self._base_client.ClearExecutionMemory(spm_pb2.Scenario(scenario=scenario, execution=execution))
+
+    def remove_data_execution(self, scenario, execution):
+        self._base_client.RemoveExecutionData(spm_pb2.Scenario(scenario=scenario, execution=execution))
+
     def update_job_by_backend(self, project, scenario, job, exe, message, status, log=None, complete=False):
         self._base_client.SetScenarioProgress(spm_pb2.ScenarioProgress(
             project=project, scenario=scenario, job_id=job,
@@ -1144,6 +1332,151 @@ class NumerousAdminClient(NumerousClient):
             yield l.log_entry, datetime.fromtimestamp(l.timestamp)
 
 
+
+    def make_commit(self, path, repository='test', branch='master', comment=None, parent_commit_id=None, dry_run=False, snapshot=True):
+
+        if comment is None or comment == '':
+            raise ValueError('Comment cannot be None or empty string.')
+
+        d = Dir(path)
+        dir_state = DirState(d, index_cmp=filehash)
+
+        state_ = dir_state.state.copy()
+        state_.pop("directory")
+
+        snapshot_id = d.hash()
+        if not dry_run:
+            if snapshot:
+                try:
+                    response = self._build_manager.CreateSnapshot(spm_pb2.Snapshot(repository=repository, snapshot_id=snapshot_id, file_structure=json.dumps(state_), only_get_uploads=False))
+
+                    def upload(filename, upload_url):
+
+
+                        filesize = os.path.getsize(filename)
+
+                        headers = {"Content-Type": "multipart/related"}
+                        params = {
+                            "name": filename,
+                            "mimeType": "application/octet-stream"
+                        }
+                        r = requests.post(
+                            upload_url,
+                            headers=headers,
+                            params=params,
+                            data=open(filename, 'rb')
+                        )
+
+                    for f, u in zip(response.filenames, response.upload_urls):
+                        log.debug('Uploading: '+path+"/"+f)
+                        upload(path+"/"+f,u)
+
+
+                except grpc._channel._InactiveRpcError as e:
+                    if not e.args[0].code == grpc.StatusCode.ALREADY_EXISTS:
+                        raise
+                    else:
+                        log.debug('Snapshot already exist. Ignoring.')
+
+                log.debug('upload complete')
+
+            histroy_reply = self._build_manager.CreateHistoryUpdate(spm_pb2.HistoryUpdate(
+                repository=repository,
+                branch=branch,
+                snapshot_id=snapshot_id,
+                parent_commit_id=parent_commit_id,
+                comment=comment,
+                timestamp=datetime.utcnow().timestamp()
+
+            ))
+            log.debug('Commit complete')
+
+            return histroy_reply.commit_id, snapshot_id, None
+        else:
+
+            return None, snapshot_id, state_
+
+    def pull_snapshot(self, repository, snapshot_id, local_path):
+        snapshot = self._build_manager.PullSnapshot(spm_pb2.PullRequest(repository=repository, snapshot_id=snapshot_id))
+        print(snapshot.subdirs)
+        for sd in snapshot.subdirs:
+            path = Path(local_path+'/'+sd)
+
+            path.mkdir(parents=True, exist_ok=True)
+
+
+
+        files_out = {}
+        for f in snapshot.files:
+            r = requests.get(f.url, allow_redirects=True)
+            local_file_path = local_path + '/' + f.name.replace('\\\\','/').replace('\\','/')
+            open(local_file_path, 'wb+').write(r.content)
+            files_out[f.name] = local_file_path
+
+        return files_out
+
+    def generate_scenario_token(self, project, scenario, job, user_id, organization_id, agent, purpose, admin=False):
+
+        token = self._token_manager.CreateRefreshToken(spm_pb2.TokenRequest(
+            project_id = project,
+            scenario_id = scenario,
+            job_id = job,
+            user_id=user_id,
+            organization_id=organization_id,
+            agent=agent,
+            purpose=purpose,
+            admin=admin
+            )).val
+
+        return token
+
+
+
+    def enqueue_build(self, repository, commit_id, launch_after_build=False):
+        resp = self._build_manager.LaunchBuild(spm_pb2.BuildRequest(repository=repository, commit_id=commit_id, launch_after_build=launch_after_build))
+        return resp.build_id
+
+    def update_scenario_image(self, name, path, project, scenario, job, repository, commit, build, internal=True):
+        self._build_manager.UpdateJobImageReference(spm_pb2.JobImageReferenceUpdate(project = project, scenario = scenario, job = job, name=name, path=path,
+                                                                                    repository=repository, commit_id=commit, internal=internal, build=build
+                                                                                    ))
+
+    def build_worker(self, work_function):
+        reponse_queue = Queue()
+
+        def responses():
+            while True:
+                yield spm_pb2.BuildResponse(status=reponse_queue.get())
+
+        for work in self._build_manager.Builder(responses()):
+            reponse_queue.put('ACK')
+            reponse_queue.put(work_function(work.json))
+
+        #response = self._build_manager.Builder(spm_pb2.BuildResponse)
+
+    def update_build(self, repository, build_id, **update):
+        self._build_manager.UpdateBuild(spm_pb2.BuildUpdate(repository=repository, build_id=build_id, update=json.dumps(update)))
+
+    def get_build_info(self, repository, build_id):
+        response = self._build_manager.GetBuildInfo(spm_pb2.BuildInfoRequest(repository=repository, build_id=build_id))
+        return json.loads(response.info)
+
+    def get_history(self, repository, branch=None, history_updates=None):
+
+        response = self._build_manager.GetHistory(spm_pb2.HistoryRequest(repository=repository, branch=branch, history_updates=history_updates))
+        return response.updates
+
+    def push_scenario_logs_admin(self, logs, exe_id):
+
+
+        timestamps = [l[0].timestamp() for l in logs]
+
+        self._base_client.PushExecutionLogEntries(spm_pb2.LogEntries(
+            scenario='',
+            execution_id=exe_id, log_entries=[l[1] for l in logs],
+            timestamps=timestamps
+        ))
+
 @contextmanager
 def open_admin_client(server=None, port=None, secure=None, refresh_token=None) -> NumerousAdminClient:
     numerous_client_ = NumerousAdminClient(server=server, port=port, secure=secure, refresh_token=refresh_token)
